"""
Represents an exception that occurs during an EYAML command execution.

Copyright 2019 William W. Kimball, Jr. MBA MSIS
"""


class EYAMLCommandException(Exception):
    """Exception to raise when an EYAML command executation fails."""
